#include "tran.h"

/*
 * Global Variables
 */
FILE *fIn;
FILE *fOut;
Pila pilaEnts;
char propietat[TAM_BUF];
int compProp = FALSE; /* Flag que indica si podem comparar la propietat */
int propCompt = 0;
int lastTagClass = FALSE; /* Indica si l'ultim tag era in_class */
char nomDir[100] = "http://cidoc.ics.forth.gr/docs/xml_to_rdfs/CIDOC_v3.3.2.rdfs#";

/*
 * Auxiliar Functions
 */

/* error: Writes an error message and returns to shell.
          If sistema=True, shows the system message
 */
void error(char *miss, int sistema)
{
    if (!sistema)
        write(STDERR, miss, strlen(miss));
    else
        perror(miss);
    getch();
    exit(1);
}

/* saltaBlancs: It is used to ignore wite spaces, tabs and carry returns */
char saltaBlancs()
{
	char car;

    fscanf(fIn, "%c", &car);
    while (car == ' ' || car == '\n' || car == '\t')
		fscanf(fIn, "%c", &car);
	return car;
}

/* trobarPunt: Finds the position inside an entity name, where the piont
               must be. It is after the entity number
 */
int trobarPunt(char nomEnt[])
{
	char *punt, *aux;
    int i = 1; /* The position 0 is the letter E */

    if ((nomEnt[0] != 'E') && (nomEnt[0] != 'e'))
    	return FALSE;

    punt = &nomEnt[i];
    while (*punt >= '0' && *punt <= '9')
    	punt = &nomEnt[++i];

    if (i < 2)
    	return FALSE;

  	*punt = '.';
    punt++;
    /* Now we eat the white spaces */
    aux = punt;
    while (*punt == ' ')
        punt++;

    while (*punt != '\0')
	{
		*aux = *punt;
     	aux++;
       	punt++;
	}
	*aux = *punt;  /* Copy the \0 */
    return TRUE;
}


/*
 * Stack Functions
 */
/* inicialitzarPila: Inicializes the stack with a NULL pointer and
   					 puts to 0 the number of elements in the stack
 */
void inicialitzarPila()
{
	pilaEnts.top = NULL;
    pilaEnts.numElems = 0;
}

/* afegirPila: Adds a new entity name to the stack, incrementing the counter */
void afegirPila(char nom[])
{
	NodePila *n;

    n = (NodePila *) malloc(sizeof(NodePila));
    strcpy(n->nomEnt, nom);
   	n->anterior = pilaEnts.top;
    pilaEnts.top = n;
   	(pilaEnts.numElems)++;
}

/* buidarPila: Pops an element from the stack, decreasing the counter */
void buidarPila()
{
	NodePila *n;

	if (pilaEnts.top == NULL)
    	error("Accessing Entity Stack without elements\n",FALSE);
    else
    {
		(pilaEnts.numElems)--;
	    n = pilaEnts.top;
	    pilaEnts.top = pilaEnts.top->anterior;
	    free(n);
    }
}

/* Compara el tag final 'tag' amb el tag inicial 'propietat' */
int comparaTags(char tag[])
{
	char *pcar;

    if (compProp)
    {
	    pcar = &tag[1];
		return(!strcmp(propietat, pcar));
    }
    else
    	return FALSE;
}

/*
 * Program Functions
 */
/* llegirTag: Reads a Tag from the Input File
              Assumes that the file pointer is in the beggining
              of a tag ('<') or in the spaces before it.
 */
void llegirTag(char tag[])
{
	char car;
    int i = 0;

    saltaBlancs();
  	fscanf(fIn, "%c", &car);
    while (car != '>')
    {
        tag[i] = car;
        i++;
        fscanf(fIn, "%c", &car);
    }
    tag[i] = '\0';
}

/* llegirEntitat: Reads the Entity name from the Input file and returns
                  the string to be writen in the Output file.
                  It assumes that the file pointer is after the tag
                  '<in_class>'
 */
char *llegirEntitat(char string[])
{
	char nomEnt[TAM_BUF];
    char res[TAM_BUF];
    char *punt;

    llegirString(nomEnt);

    if (trobarPunt(nomEnt) == TRUE)
	    if (!lastTagClass)
    	{
       		sprintf(res, "<crm:%s rdf:about=\"%s\">\n", nomEnt, string);
			afegirPila(nomEnt);
    	    lastTagClass = TRUE;
	    }
    	else
	    {
    	   	sprintf(res, "<rdf:type rdf:resource=\"%s%s\"/>\n",nomDir, nomEnt);
	    }
    else
    {
    	sprintf(res,"Error in entity name %s\n",nomEnt);
    	error(res,FALSE);
    }

    llegirTag(nomEnt); /* El '/in_class' */
    return res;
}

/* llegirNomProp: It reads the information of a property and writes it
                  in the rdf-crm format.
                  It assumes that we have read a tag diferent from
                  '<in_class>'
 */
char *llegirNomProp(char tag[], char string[])
{
    char *punt;
    char res1[TAM_BUF];  /* Auxiliar string */
    char res2[TAM_BUF];  /* Output string */

	if (!(strcmp(tag, "/CRM_Entity")))
    {
		sprintf(res2, "</crm:%s>\n", pilaEnts.top->nomEnt);
         buidarPila();
    }
	else
    {
	/* Change '_' for '.' */
    	if ((punt = strstr(tag, "_")) != NULL)
    		*punt = '.';
	    else
        {
        	sprintf(res1, "Missing '_' in property name %s\n", tag);
    		error(res1, FALSE);
        }

		/* There is no entity name after the property */
        if (compProp)
        {
  	    	/* Cal tornar enrera 3 posicions ('\n' i '>') */
			if (fseek(fOut, -3L, SEEK_CUR))
				error("Error fseek ", TRUE);
	   	    sprintf(res1, " rdf:resource=\"%s\">\n", string);
        }
        else
        	sprintf(res1, "");

        /* Range is literal */
        if (((punt = strstr(tag, "has_note")) != NULL) ||
    	   ((punt = strstr(tag, "has_number_of_parts")) != NULL) ||
           ((punt = strstr(tag, "begins_at_qualify")) != NULL) ||
           ((punt = strstr(tag, "ends_at_qualify")) != NULL) ||
           ((punt = strstr(tag, "has_value")) != NULL) ||
           ((punt = strstr(tag, "at_some_time_within")) != NULL) ||
           ((punt = strstr(tag, "ongoing_throughout")) != NULL)
         )
        {
        	llegirString(string);
            sprintf(res2, "%s<crm:%s>%s</crm:%s>\n",res1,tag,string,tag);
            llegirTag(tag); /* el has_note */
            compProp = FALSE;
        }
        else
        {
		    if (tag[0] != '/')	/* Initial tag */
    		{
        		llegirString(string);
	        	strcpy(propietat, tag);
	    	    sprintf(res2, "%s<crm:%s>\n",res1,tag);
    	    	propCompt++;
		        compProp = TRUE;
		    }
    		else
	    	{
    	    	punt = &tag[1]; /* Saltem el '/' */
        	    if (comparaTags(tag)) /* Tag final consecutiu */
            	{
	        	    sprintf(res1, " rdf:resource=\"%s\"/>\n", string);
		        }
    	        else
			      	sprintf(res1, "%s</crm:%s>\n", res1, punt);

	        	if (propCompt < pilaEnts.numElems && !compProp) /* Imprimim tag final entitat */
   		        {
       				sprintf(res2, "</crm:%s>\n%s",pilaEnts.top->nomEnt, res1);
           		    buidarPila();
            	}
		        else
					sprintf(res2, "%s", res1);

                compProp = FALSE;
                propCompt--;
		    }
        }
    }
    return res2;
}

/* llegirString: Reads characters from the input file until it reaches
                 the character '<'
 */
void llegirString(char string[])
{
	char car;
    int i = 1;

    string[0] = saltaBlancs();
    fscanf(fIn, "%c", &car);
    while (car != '<')
    {
        string[i] = car;
        i++;
        fscanf(fIn, "%c", &car);
    }
    string[i] = '\0';

    /* Ens menjem el \n final */
    for (i = i-1; i>0; i--)
    	if (string[i] == ' ' || string[i] == '\t')
			continue;
        else
        	if (string[i] == '\n')
            {
            	string[i] = '\0';
                break;
            }
            else
            	break;

	/* Tornem una posicio enrera en el fitxer
       per poder llegir el seguent tag        */
    if (fseek(fIn, -1L, SEEK_CUR))
		error("Error fseek ", TRUE);
}

/* escriureRes: Writes a string to the Output File */
void escriureRes(char string[])
{
	fprintf(fOut, "%s", string);
}


/*
 * Main function
 */
void main(int argc, char **argv)
{
	char string[TAM_BUF];
    char tag[TAM_BUF];

    if (argc != 3)
    {
        sprintf(string, "Utilitzacio: %s fileIn fileOut\n", argv[0]);
        error(string, FALSE);
    }

	/* Inizialization of the stack */
    inicialitzarPila();

    /* Open Input File */
    if ((fIn = fopen( argv[1],"r")) == NULL)
    {
        sprintf(string, "Error in opening input file %s\n", argv[1]);
        error(string, TRUE);
    }

    /* Open Output File */
    if ((fOut = fopen( argv[2],"w")) == NULL)
    {
        sprintf(string, "Error in opening output file %s\n", argv[2]);
        error(string, TRUE);
    }

    /* Read xml Tags */
    /* Read CRMset tag */
    llegirTag(tag);
    while (strcmp(tag, "CRMset"))
		llegirTag(tag);

    /* Write RDF headings */
    fprintf(fOut, HEAD1);
    fprintf(fOut, HEAD2);

    /* Main Loop */
    /* Read first tag (CRM_Entity) */
    llegirTag(tag);
    while (strcmp(tag, "/CRMset"))
    {
		if (!(strcmp(tag, "CRM_Entity")))
    		llegirString(string);
    	else
	        if (!(strcmp(tag, "in_class")))
            {
                compProp = FALSE;
           	    escriureRes(llegirEntitat(string));
            }
   	    	else
            {
                escriureRes(llegirNomProp(tag, string));
                lastTagClass = FALSE;
            }
       	llegirTag(tag);
    }

    /* Write RDF ending */
    fprintf(fOut, "</rdf:RDF>");
    /* Close Input and Output Files */
    fclose(fIn);
    fclose(fOut);
}